@@ -30,6 +30,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/poll.h>
+#include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -38,6 +39,7 @@
 #include "execute.h"
 #include "log.h"
 #include "target.h"
+#include "util.h"
 
 const char *const execute_environment_target[] =
 {
@@ -46,11 +48,8 @@ const char *const execute_environment_target[] =
   NULL
 };
 
-const struct execute_io_info execute_io_log_info[] = {
-  { 1, execute_io_log_handler, (void *)DI_LOG_LEVEL_OUTPUT },
-  { 2, execute_io_log_handler, (void *)LOG_LEVEL_OUTPUT_STDERR },
-};
-const unsigned int execute_io_log_info_count = 2;
+const struct execute_io_info execute_io_log_info[] = { EXECUTE_IO_LOG };
+const unsigned int execute_io_log_info_count = NELEMS (execute_io_log_info);
 
 static int internal_di_exec_child (const char *const argv[], const char *const envp[], pid_t pid, di_process_handler *child_prepare_handler, void *child_prepare_user_data, int fd_status)
 {
@@ -79,6 +78,8 @@ static int internal_status_handle (int fd)
 
   while (poll (pollfds, 1, -1) >= 0)
   {
+    if (pollfds[0].revents & POLLHUP)
+      return 0;
     if (pollfds[0].revents & POLLIN)
     {
       int status;
@@ -86,8 +87,6 @@ static int internal_status_handle (int fd)
         abort ();
       return status;
     }
-    if (pollfds[0].revents & POLLHUP)
-      return 0;
   }
 
   abort ();
@@ -101,75 +100,69 @@ static void internal_io_handle (int fds[][2], const struct execute_io_info io_in
   for (unsigned int i = 0; i < io_info_count; ++i)
   {
     pollfds[i].fd = fds[i][0];
-    pollfds[i].events = POLLIN;
-    files[i] = fdopen (fds[i][0], "r");
+    pollfds[i].events = 0;
+    files[i] = NULL;
+
+    if (io_info[i].handler)
+    {
+      pollfds[i].events = io_info[i].events;
+      if (io_info[i].events == POLLIN)
+        files[i] = fdopen (fds[i][0], "r");
+      else if (io_info[i].events == POLLOUT)
+        files[i] = fdopen (fds[i][0], "w");
+    }
   }
 
   while (poll (pollfds, io_info_count, -1) >= 0)
   {
-    bool exit = false;
-
     for (unsigned int i = 0; i < io_info_count; ++i)
-    {
-      if (pollfds[i].revents & POLLIN)
-      {
-        char line[1024];
-        while (fgets (line, sizeof (line), files[i]) != NULL)
-          io_info[i].handler (line, strlen (line), io_info[i].user_data);
-        exit = true;
-      }
-    }
-
-    if (exit)
-      continue;
+      if (pollfds[i].revents & (POLLIN | POLLOUT))
+        io_info[i].handler (files[i], io_info[i].user_data);
 
     for (unsigned int i = 0; i < io_info_count; ++i)
       if (pollfds[i].revents & POLLHUP)
-        exit = true;
-
-    if (exit)
-      break;
+        // XXX
+        goto cleanup;
   }
 
+cleanup:
   for (unsigned int i = 0; i < io_info_count; ++i)
-    fclose (files[i]);
+    if (files[i])
+      fclose (files[i]);
 }
 
 static int internal_di_exec (const char *const argv[], const char *const envp[], const struct execute_io_info io_info[], unsigned int io_info_count, di_process_handler *child_prepare_handler, void *child_prepare_user_data)
 {
   pid_t pid;
   int fd_null, fds_status[2], fds_io[io_info_count][2];
-  int errno_saved;
   int ret = 0;
 
-  fd_null = open ("/dev/null", O_RDWR);
+  fd_null = open ("/dev/null", O_RDWR | O_CLOEXEC);
 
-  if (pipe (fds_status) < 0)
+  if (pipe2 (fds_status, O_CLOEXEC) < 0)
     abort ();
-  fcntl (fds_status[1], F_SETFD, FD_CLOEXEC);
 
   for (unsigned int i = 0; i < io_info_count; ++i)
   {
-    if (pipe (fds_io[i]) < 0)
-      abort ();
-    fcntl (fds_io[i][0], F_SETFD, FD_CLOEXEC);
-    fcntl (fds_io[i][0], F_SETFL, O_NONBLOCK);
+    if (io_info[i].handler)
+    {
+      if (socketpair (AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, fds_io[i]) < 0)
+        abort ();
+      if (io_info[i].events & POLLIN)
+        fcntl (fds_io[i][0], F_SETFL, O_NONBLOCK);
+    }
+    else
+    {
+      fds_io[i][0] = dup (fd_null);
+      fds_io[i][1] = fd_null;
+    }
   }
 
-  pid = fork ();
-  errno_saved = errno;
-
-  // XXX: Flow?
-  if (pid <= 0)
-  {
-    close (fds_status[0]);
-    for (unsigned int i = 0; i < io_info_count; ++i)
-      close (fds_io[i][0]);
-  }
+  if ((pid = fork ()) < 0)
+    abort ();
 
   if (pid == 0)
   {
-    dup2 (fd_null, 0);
     for (unsigned int i = 0; i < io_info_count; ++i)
       dup2 (fds_io[i][1], io_info[i].fd);
 
@@ -177,19 +170,10 @@ static int internal_di_exec (const char *const argv[], const char *const envp[],
   }
 
   close (fd_null);
-
-  if (pid < 0)
-  {
-    ret = -errno_saved;
-    goto out;
-  }
-
   close (fds_status[1]);
   for (unsigned int i = 0; i < io_info_count; ++i)
     close (fds_io[i][1]);
 
-  errno_saved = -1;
-
   if ((ret = internal_status_handle (fds_status[0])))
     goto out;
 
@@ -210,12 +194,16 @@ out:
   return ret;
 }
 
-int execute_io_log_handler (char *buf, size_t n __attribute__ ((unused)), void *user_data)
+void execute_io_log_handler (FILE *f, void *user_data)
 { 
-  if (buf[n - 1] == '\n')
-    buf[n - 1] = 0;
-  log_text ((uintptr_t)user_data, "%s", buf);
-  return 0;
+  char buf[1024];
+  while (fgets (buf, sizeof (buf), f))
+  {
+    size_t n = strlen (buf);
+    if (buf[n - 1] == '\n')
+      buf[n - 1] = 0;
+    log_text ((uintptr_t)user_data, "%s", buf);
+  }
 }
 
 static void execute_status (int status)
