@@ -389,9 +389,11 @@ static int foreign_finish ()
     return ret;
 
   fputs ("\
-rm /sbin/init\n\
-dpkg-divert --remove --local --rename --divert /sbin/init.REAL /sbin/init\n\
-rm /sbin/cdebootstrap-foreign\n\
+run rm /sbin/init\n\
+run dpkg-divert --remove --local --rename --divert /sbin/init.REAL /sbin/init\n\
+run rm /sbin/cdebootstrap-foreign\n\
+\n\
+finish\n\
 ", foreign_script);
 
   log_text (DI_LOG_LEVEL_INFO, "Second stage installer is available as /sbin/cdebootstrap-foreign or /sbin/init");
@@ -402,6 +404,7 @@ rm /sbin/cdebootstrap-foreign\n\
 static install_simulate_handler foreign_handler;
 static int foreign_handler (const char *const command[])
 {
+  fputs ("run ", foreign_script);
   for (; *command; command++)
   {
     fputs (*command, foreign_script);
@@ -431,8 +434,6 @@ static int foreign_init ()
   fputs ("\
 #!/bin/sh\n\
 \n\
-set -e\n\
-\n\
 ", foreign_script);
 
   const char *const *temp = execute_environment_target;
@@ -441,20 +442,40 @@ set -e\n\
 
   fputs ("\
 \n\
-if [ \"$0\" = /sbin/init ]; then\n\
-  trap 'exec env -u DEBIAN_FRONTEND /bin/sh' ERR\n\
-  trap 'exec env -u DEBIAN_FRONTEND -u PATH /sbin/init' EXIT\n\
-  mount -n -o remount,rw rootfs /\n\
-else\n\
-  trap 'umount /proc || :' EXIT\n\
-  mount -n -t proc none /proc\n\
-fi\n\
+exec_pid1() {\n\
+  [ \"$$\" = 1 ] && exec env -u DEBIAN_FRONTEND \"$@\" || :\n\
+}\n\
+\n\
+finish() {\n\
+  exec_pid1 /sbin/init\n\
+  umount /proc || :\n\
+  exit 0\n\
+}\n\
+\n\
+error() {\n\
+  echo \"$1\" 2>&1\n\
+  exec_pid1 /bin/sh\n\
+  umount /proc || :\n\
+  exit 1\n\
+}\n\
+\n\
+run() {\n\
+  \"$@\" 3>/dev/null || error \"Failed to run $1!\"\n\
+}\n\
+\n\
+trap 'error \"Interruped!\"' HUP INT TERM\n\
+\n\
+mount -n -o remount,rw rootfs /\n\
+\n\
+chown -hR 0:0 /\n\
+\n\
+mount -n -t proc none /proc\n\
 \n\
 if [ ! -e /sbin/init.REAL ]; then\n\
-  rm /sbin/init\n\
-  ln -s /bin/sh /sbin/init\n\
-  dpkg-divert --add --local --divert /sbin/init.REAL /sbin/init\n\
-  mv /sbin/init.foreign /sbin/init.REAL\n\
+  run rm /sbin/init\n\
+  run ln -s /bin/sh /sbin/init\n\
+  run dpkg-divert --add --local --divert /sbin/init.REAL /sbin/init\n\
+  run mv /sbin/init.foreign /sbin/init.REAL\n\
 fi\n\
 \n\
 ", foreign_script);
@@ -676,20 +697,20 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
   if (foreign && foreign_init ())
     log_text (DI_LOG_LEVEL_ERROR, "Internal error: foreign init");
 
-  if (!download_only)
-  {
-    install_simulate_handler *tmp = NULL;
-    if (foreign)
-      tmp = foreign_handler;
-    if (install_init (helperdir, tmp))
-      log_text (DI_LOG_LEVEL_ERROR, "Internal error: install init");
-  }
-
   if (download (&packages, &allocator, &list))
     log_text (DI_LOG_LEVEL_ERROR, "Internal error: download");
 
   if (download_only)
+  {
+    log_text (DI_LOG_LEVEL_INFO, "Download-only mode, not installing anything");
     return 0;
+  }
+
+  install_simulate_handler *tmp = NULL;
+  if (foreign)
+    tmp = foreign_handler;
+  if (install_init (helperdir, tmp))
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: install init");
 
   if (install (packages, allocator, list))
     log_text (DI_LOG_LEVEL_ERROR, "Internal error: install");
