@@ -31,6 +31,7 @@
 #include "gpg.h"
 #include "log.h"
 #include "suite.h"
+#include "util.h"
 
 static const char *keyring = NULL;
 static const char **keyringdirs;
@@ -44,31 +45,32 @@ struct check_release
 const char GNUPGBADSIG[] = "[GNUPG:] BADSIG";
 const char GNUPGGOODSIG[] = "[GNUPG:] GOODSIG";
 
-static int check_release_status_io_handler (char *buf, size_t n __attribute__ ((unused)), void *user_data)
+static void check_release_status_io_handler (FILE *f, void *user_data)
 { 
   struct check_release *data = user_data;
+  char buf[1024];
 
-  if (buf[n - 1] == '\n')
+  while (fgets (buf, sizeof (buf), f))
   {
-    buf[n - 1] = 0;
-    n--;
-  }
+    size_t n = strlen (buf);
+    if (buf[n - 1] == '\n')
+      buf[n - 1] = 0;
 
-  log_text (DI_LOG_LEVEL_DEBUG, "Status \"%s\"", buf);
+    log_text (DI_LOG_LEVEL_DEBUG, "GNUPG Status: %s", buf);
 
-  if (strncmp (buf, GNUPGGOODSIG, strlen (GNUPGGOODSIG)) == 0)
-  {
-    const char *b = buf + strlen (GNUPGGOODSIG) + 16 + 2;
-    data->goodsigs++;
-    log_text (DI_LOG_LEVEL_INFO, "Good signature from \"%s\"", b);
+    if (strncmp (buf, GNUPGGOODSIG, strlen (GNUPGGOODSIG)) == 0)
+    {
+      const char *b = buf + strlen (GNUPGGOODSIG) + 16 + 2;
+      data->goodsigs++;
+      log_text (DI_LOG_LEVEL_INFO, "Good signature from \"%s\"", b);
+    }
+    else if (strncmp (buf, GNUPGBADSIG, strlen (GNUPGBADSIG)) == 0)
+    {
+      const char *b = buf + strlen (GNUPGBADSIG) + 16 + 2;
+      data->badsigs++;
+      log_text (DI_LOG_LEVEL_WARNING, "BAD signature from \"%s\"", b);
+    }
   }
-  else if (strncmp (buf, GNUPGBADSIG, strlen (GNUPGBADSIG)) == 0)
-  {
-    const char *b = buf + strlen (GNUPGBADSIG) + 16 + 2;
-    data->badsigs++;
-    log_text (DI_LOG_LEVEL_WARNING, "BAD signature from \"%s\"", b);
-  }
-  return 0;
 }
 
 int gpg_check_release (const char *file, const char *file_sig, const char *message)
@@ -82,9 +84,8 @@ int gpg_check_release (const char *file, const char *file_sig, const char *messa
     NULL, NULL, NULL,
   };
   struct execute_io_info io_info[] = {
-    { 1, execute_io_log_handler, (void *)DI_LOG_LEVEL_OUTPUT },
-    { 2, execute_io_log_handler, (void *)LOG_LEVEL_OUTPUT_STDERR },
-    { 3, check_release_status_io_handler, &data },
+    EXECUTE_IO_LOG,
+    { 3, POLLIN, check_release_status_io_handler, &data },
   };
   
   if (!keyring)
@@ -100,7 +101,7 @@ int gpg_check_release (const char *file, const char *file_sig, const char *messa
   
   log_message (LOG_MESSAGE_INFO_DOWNLOAD_VALIDATE, message);
 
-  execute_full (command, io_info, 3);
+  execute_full (command, io_info, NELEMS (io_info));
 
   return !data.goodsigs || data.badsigs;
 }
